// IndentedTrace.cpp: implementation of the CIndentedTrace class.
//
// Better tracing class and macros
// 
// Copyright (C) 2000 Mark Messer - Released to the public domain
// This software is provided "as is" without warranty of any kind. 
// 
// Revision History
// 11/28/00	MM	Initial release
// 
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "IndentedTrace.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
	// Add this class to a program to print a trace of function calls in the 
	// output window for debug compiles. For non-debug compiles, the class 
	// contains only empty functions. Tracing must be enabled for this to work.
	// 
	// Declare a CIndentedTrace variable at the beginning of each function 
	// to be traced. Give the class and function names as the argument. The 
	// IndentedTrace constructor will print the class and function names, 
	// and an opening brace in the output window and increment the indent 
	// level. 
	//
	// Example of use:
	// 
	// CMyClass::MyFunc()
	// {
	//		CIndentedTrace IT("CMyClass::MyFunc()");
	//		...
	// }
	// 
	// Traced functions called by the first function are printed at the new 
	// indent level. 
	// 
	// The variable goes out of scope at the end of the function. The 
	// destructor decrements the indent and prints a closing brace. 
	// 
	// Some functions contain many repetitive function calls. To suppress 
	// CIndentedTrace tracing inside such functions, call SuppressTrace(). 
	// Tracing will resume when the function exits.
	// 
	// To print a comment at the current indent level, call Comment().


//////////////////////////////////////////////////////////////////////
	// Initialize static data members.

#ifdef _DEBUG

	// Set initial indent level to no indentation. 
int CIndentedTrace::ms_iTraceDepth = 0;		

	// Deactivate trace suppression. See SuppressTrace() for more info.  
int CIndentedTrace::ms_iSuppressDepth = 0;	

	// Initialize to 0 so first serial number will be 1. 
int CIndentedTrace::ms_iSerialNumber = 0;

	// By default, do not display serial numbers.
bool CIndentedTrace::ms_bEnableSerialNumber = false;

	// By default, do not display thread ID.
bool CIndentedTrace::ms_bEnableThreadID = false;
 
#endif // _DEBUG

//////////////////////////////////////////////////////////////////////
// Implement private helper functions. 

// ---------------------------------------------------------------------------
// AppendSN()
// 
// Appends the serial number of this instance to strText (unless modified by 
// trace suppression or relase compilation). 
// 
// strText - IN. String to which the serial number will be appended. 
// 
// Each time a CIndentedTrace object is constructed, the constructor reserves 
// a serial number for object. Serial numbers start at 1, and are incremented 
// each time the constructor is called. 
// 
// This function is only called inside CIndentedTrace functions that are 
// changed when tracing is suppressed. 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::AppendSN(CString& strText)
{
#ifdef _DEBUG

	if (ms_bEnableSerialNumber)
	{
		CString strLSN;

		strLSN.Format("\t\t// S/N %d", m_nLocalSerialNumber);
		strText += strLSN;
	}

#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// AppendThdId()
// 
// Appends the ID of the thread this instance runs in to strText, unless it 
// is the main thread (or unless modified by trace suppression or relase 
// compilation).  
// 
// strText - IN. String to which the serial number will be appended. 
// 
// This function is only called inside CIndentedTrace functions that are 
// changed when tracing is suppressed. 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::AppendThdId(CString& strText)
{
#ifdef _DEBUG
	if (ms_bEnableThreadID)
	{
		if (AfxGetThread()->m_hThread != AfxGetApp()->m_hThread)
		{
			CString strThdId;

			strThdId.Format("\t\t// Thd ID %X", AfxGetThread()->m_nThreadID);
			strText += strThdId;
		}
	}
#endif // _DEBUG
}


//////////////////////////////////////////////////////////////////////
// Constructors/Destructors

// ---------------------------------------------------------------------------
// CIndentedTrace( LPCTSTR )
// 
// This constructor assigns this instance a serial number, prints an indented 
// line in the output window with opening brace, and increments the indent 
// level (unless modified by trace suppression or relase compilation). 
// 
// strClass_Func - IN. String that is printed. 
// 
// strClass_Func usually contains the name of the class and function that 
// calls this constructor. E.g. "CMyClass::MyFunc()".
// 
// When strClass_Func is printed in the output window, spaces are added to the 
// beginning to indent it, and an opening brace is added to the end. Depending 
// on ms_bEnableSerialNumber and ms_bEnableThreadID, a serial number and/or 
// thread ID may be added to the end as well. 
// 
// The static member variable ms_iSerialNumber contains the most recently used 
// taken serial number. The serial number is reserved by copying it and 
// incrementing ms_iSerialNumber. 
// 
// ---------------------------------------------------------------------------
CIndentedTrace::CIndentedTrace(LPCTSTR strClass_Func)
{
#ifdef _DEBUG

		m_nLocalSerialNumber = ++ms_iSerialNumber;
		Entry(strClass_Func);

#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// ~CIndentedTrace()
// 
// The destructor decrements the indent level and prints an indented closing 
// brace (unless modified by trace suppression or relase compilation).
// 
// 
// ---------------------------------------------------------------------------
CIndentedTrace::~CIndentedTrace()
{
#ifdef _DEBUG

		Exit();

#endif // _DEBUG
}


//////////////////////////////////////////////////////////////////////

// ---------------------------------------------------------------------------
// Entry()
// 
// Prints an indented line in the output window with opening brace, and 
// increments the indent level (unless trace suppression is active or this is 
// a relase compile). 
// 
// strClass_Func - IN. String that is printed. 
// 
// strClass_Func usually contains the name of the class and function that 
// created this instance of CIndentedTrace. E.g. "CMyClass::MyFunc()".
// 
// When strClass_Func is printed in the output window, spaces are added to the 
// beginning to indent it, and an opening brace is added to the end. Depending 
// on ms_bEnableSerialNumber and ms_bEnableThreadID, a serial number and/or 
// thread ID may be added to the end as well. 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::Entry(LPCTSTR strClass_Func)
{
#ifdef _DEBUG

	// Sanity checks.
	ASSERT (ms_iTraceDepth >= 0);
	ASSERT ((ms_iTraceDepth >= ms_iSuppressDepth) || (ms_iSuppressDepth == 0));

	m_nLocalTraceDepth = ms_iTraceDepth;

	if (ms_iSuppressDepth == 0)
	{		// Suppression inactive. Do trace. 
		CString strIndent(' ', INDENT_SPACES * m_nLocalTraceDepth);
		CString strText;

		strText = strIndent + strClass_Func + " {";

		AppendThdId(strText); // Do nothing if not enabled. 
		AppendSN(strText);	  // Do nothing if not enabled. 
		strText += '\n';
		TRACE (strText);
	}

	++ms_iTraceDepth;
	++m_nLocalTraceDepth;

#else // _DEBUG

	UNUSED(strClass_Func);	// Prevent compiler warning

#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// Exit()
// 
// Opposite of Entry(). Decrements the indent level and prints an indented 
// closing brace (unless trace suppression is active or this is a relase 
// compile). 
// 
// Exit() is also responsible for turning off trace suppression. For more 
// info, see SuppressTrace(). 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::Exit()
{
#ifdef _DEBUG

	// Sanity checks.
	ASSERT (ms_iTraceDepth >= 0);
	ASSERT ((ms_iTraceDepth >= ms_iSuppressDepth) || (ms_iSuppressDepth == 0));

	--ms_iTraceDepth;
	--m_nLocalTraceDepth;

	if (ms_iTraceDepth < ms_iSuppressDepth)
	{		// Exiting function that started suppression. Deactivate it.
		ms_iSuppressDepth = 0;
	}
	
	if (ms_iSuppressDepth == 0)
	{		// Suppression inactive. Do trace. 
		CString strIndent(' ', INDENT_SPACES * m_nLocalTraceDepth);
		CString strText;

		strText = strIndent + "}";

		AppendThdId(strText); // Do nothing if not enabled. 
		AppendSN(strText);	  // Do nothing if not enabled. 
		strText += '\n';
		TRACE (strText);

		if (m_nLocalTraceDepth == 0)
		{		// Insert extra line to separate blocks
			TRACE ("\n");
		}
	}

#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// Comment()
// 
// Prints an indented comment line in the output window (unless trace 
// suppression is active or this is a relase compile). 
// 
// strComment - IN. String that is printed. 
// 
// When strComment is printed in the output window, spaces are added to the 
// beginning to indent it, and "// " is added to the beginning. Depending 
// on ms_bEnableSerialNumber and ms_bEnableThreadID, a serial number and/or 
// thread ID may be added to the end as well. 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::Comment(LPCTSTR strComment)
{
		// Print an indented comment. 

#ifdef _DEBUG
	if (ms_iSuppressDepth == 0)
	{
		CString strIndent(' ', INDENT_SPACES * m_nLocalTraceDepth);
		strIndent += "// "; 
		strIndent += strComment; 
		AppendThdId(strIndent);		// Do nothing if not enabled. 
		AppendSN(strIndent);		// Do nothing if not enabled. 
		strIndent += "\n";
		TRACE (strIndent);
	}

#else // _DEBUG
	UNUSED(strComment);				// Prevent compiler warning.
#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// SuppressTrace()
// 
// Suppresses action of CIndentedTrace functions until the the instance where 
// SuppressTrace() is called exits (unless this is a release compile). 
// 
// This is function was intended to prevent functions that produce a lot of 
// CIndentedTrace output from cluttering the output window. It turns out not 
// to be very useful. 
// 
// Each CIndentedTrace function that produces output checks ms_iSuppressDepth. 
// If ms_iSuppressDepth > 0, suppression is active, and the CIndentedTrace 
// function does little or nothing. 
// 
// Normally, ms_iSuppressDepth is 0, which disables suppression. 
// 
// Suppose a function, MyFunc(), creates an instance of CIndentedTrace and 
// calls SuppressTrace(). This sets ms_iSuppressDepth to the current indent 
// level, starting suppression. 
// 
// If MyFunc() calls more functions, the indent level may increase. But it 
// should never drop below the current level until MyFunc() exits. 
// 
// Exit() checks for when the indent level drops below the value stored in 
// ms_iSuppressDepth. When this happens, Exit() sets ms_iSuppressDepth to 0, 
// turning off suppression. 
// 
// 
// SuppressTrace() and multithreaded apps do not mix. 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::SuppressTrace()
{
	// CIndentedTrace functions will do no further tracing until the function 
	// that called SuppressTrace() exits.
	//

#ifdef _DEBUG
	ASSERT (m_nLocalTraceDepth > 0);

	if (ms_iSuppressDepth == 0)
	{		// Start trace suppression only if it is not already started.
		ms_iSuppressDepth = m_nLocalTraceDepth; 

			// Show suppression is starting.
		CString strIndent(' ', INDENT_SPACES * m_nLocalTraceDepth);
		TRACE (strIndent + 
			"// CIndentedTrace is suppressed for the rest of this function\n");
	}

#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// DumpStr() 
// 
// Dumps the contents of a string or buffer as ASCII characters and as hex 
// numbers (unless trace suppression is active or this is a relase compile). 
// 
// str - IN. string or buffer to be dumped. 
// 
// NChars - IN. Number of chars in str. 
// 
// CharsPerLine - IN. Output may take multiple lines in the output window. 
//					This is the max number of chars from str displayed in each 
//					line. The default is 16 chars. 
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::DumpStr(LPCTSTR str, const UINT NChars, 
											const UINT CharsPerLine /* = 16 */)
{
#ifdef _DEBUG

	if (ms_iSuppressDepth != 0)
	{
		// Tracing is suppressed.
		return;
	}

	CString strIndent(' ', INDENT_SPACES * m_nLocalTraceDepth);

	CString strScr;					// Scratch
	CString strHex;					// Chars as hex numbers
	CString strAsc;					// Chars as ASCII

	for (unsigned int i=0; i<NChars; ++i)
	{
		// Add a char to formatted strings. 
		strScr.Format("%02X ", (unsigned char) str[i]);
		strHex += strScr;
		strAsc += isprint((int) str[i]) ? str[i] : '.';

		// When CharsPerLine chars have been read (after last char has 
		// been read, print formatted strings.
		if (((i+1)%CharsPerLine==0) || (i==NChars-1))
		{
			if ((unsigned int) strHex.GetLength() < 3*CharsPerLine)
			{	// Not enough chars in hex string. Pad with spaces.
				CString strScr(' ', (3*CharsPerLine) - strHex.GetLength());
				strHex += strScr;
			}

			// Combine into one string.
			strHex += "    ";
			strHex += strAsc;

			// Print in the debug output window. 
			strHex += "\n";
			TRACE(strIndent);
			TRACE(strHex);

			// Erase formatted strings that have been printed.
			strHex.Empty();
			strAsc.Empty();
		}
	}

#endif // _DEBUG
}


// ---------------------------------------------------------------------------
// TraceIOErr()
// 
// Many IO functions return errors through ::GetLastError(). This will obtain 
// and display in the output window the string that matches the error from the 
// operating system (unless trace suppression is active or this is a relase 
// compile). 
//
// GetLastError() will be left as is. 
// 
// For more info, see help on FormatMessage() or GetLastError().
// 
// ---------------------------------------------------------------------------
void CIndentedTrace::TraceIOErr()
{
#ifdef _DEBUG

	if (ms_iSuppressDepth != 0)
	{
		// Tracing is suppressed.
		return;
	}

	LPVOID lpMsgBuf;		// Ptr to buffer (allocated by FormatMessage()
	CString ErrMsg;

	FormatMessage
	( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, 
		NULL, 
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf, 
		0, 
		NULL 
	);

	ErrMsg.Format("GetLastError() reports %d - ", GetLastError());
	ErrMsg += (LPCTSTR) lpMsgBuf;
	Comment(ErrMsg);

	LocalFree( lpMsgBuf );

#endif // _DEBUG
}


